
#include <memory.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>


void hexdump_offset(
    const uint8_t *buffer,
    size_t nbytes,
    uint32_t *start,
    uint8_t *hexbuffer,
    size_t hexbuf_len,
    uint32_t breakpoint)
{
    const char *hex = "0123456789ABCDEF";
    uint32_t i, count, offset, hl;
    const uint8_t *p;
    uint8_t *hbp;
    char ascbuf[64];
    char hexbuf[64];
    char *ap = ascbuf;
    char *hp = hexbuf;

    if (!buffer || !nbytes || !hexbuffer)
        return;

    p = buffer;
    hbp = hexbuffer;
    hl = hexbuf_len;
    offset = *start;
    *ap = '\0';
    *hp = '\0';
    count = 0;
    
    if (breakpoint == 0)
    {
        breakpoint = 16;
    }
    else if (breakpoint > 32)
    {
        breakpoint = 32;
    }
    else
    {
        breakpoint &= 0xFFFFFFFE;
    }

    for (i = 0; i < nbytes; ++i)
    {
        if (i && (i & (breakpoint - 1)) == 0)
        {
            int printed = snprintf((char *)hbp, hl, "%06X %s  %s",
                offset, hexbuf, ascbuf);
            hbp += printed;
            hl -= printed;
            ap = ascbuf;
            hp = hexbuf;
            count = 0;
            offset += breakpoint;
        }
        ap += sprintf(ap, "%c", (*p >= 0x20 && *p < 0x7F) ? *p : '.');
        hp += sprintf(hp, " %c%c", hex[(*p >> 4) & 0x0F], hex[(*p & 0x0F)]);
        ++p;
        ++count;
    }

    if (count)
    {
        /* FIXME: account for breakpoint in "-48" specifier */
        snprintf((char *)hbp, hl, "%06X %-48s  %s", offset, hexbuf, ascbuf);
        offset += count;
    }
    *start = offset;
}

void hexdump(
    const void *buffer,
    size_t len,
    void (*print_fcn)(const char *, size_t, void *),
    void *print_arg)
{
    (void)print_fcn;
    uint8_t xbuf[256]; /* must be at least 5x length chunk */
    uint32_t len_chunk;
    const uint8_t *p;
    uint32_t off, l;

    p = buffer;
    off = 0;
    len_chunk = 16;
    do {
        l = len >= len_chunk ? len_chunk : len;
        hexdump_offset(p, l, &off, xbuf, sizeof(xbuf), len_chunk);
        if (print_fcn)
        {
            print_fcn((const char *)xbuf, sizeof(xbuf), print_arg);
        }
        p += l;
        len -= l;
    } while (len > 0);
}

static void hexdump_printlog(const char *buf, size_t len, void *unused)
{
    (void)len;
    (void)unused;
    syslog(LOG_INFO, "%s", buf);
}

void hexdump2log(const void *buffer, size_t len)
{
    hexdump(buffer, len, hexdump_printlog, NULL);
}
